/* $Id: interp.c,v 1.19 2009/02/19 14:42:10 mit-sato Exp $ */

#include <assert.h>
#include <string.h>
#include <signal.h>
#include "types.h"
#include "cell.h"
#include "interp.h"
#include "config.h"
#include "bulk.h"
#include "parser.h"
#include "toy.h"

Toy_Type *const_Symbol;
Toy_Type *const_Block;
Toy_Type *const_List;
Toy_Type *const_Real;
Toy_Type *const_Integer;
Toy_Type *const_String;
Toy_Type *const_Exception;
Toy_Type *const_RQuote;
Toy_Type *const_Object;
Toy_Type *const_Nil;
Toy_Type *const_T;
Toy_Type *const_Question;
Toy_Type *const_Holder;
Toy_Type *const_Init;
Toy_Type *const_Hash;
Toy_Type *const_File;
Toy_Type *const_Args;
Toy_Type *const_Array;
Toy_Type *const_Get;
Toy_Type *const_atout;
Toy_Type *const_atin;
Toy_Type *const_stdin;
Toy_Type *const_stdout;
Toy_Type *const_stderr;
Toy_Type *const_filec;
Toy_Type *const_gets;
Toy_Type *const_puts;
Toy_Type *const_atname;
Toy_Type *const_atstacktrace;
Toy_Type *const_SIGHUP;
Toy_Type *const_SIGINT;
Toy_Type *const_SIGQUIT;
Toy_Type *const_SIGPIPE;
Toy_Type *const_SIGALRM;
Toy_Type *const_SIGTERM;
Toy_Type *const_SIGURG;
Toy_Type *const_SIGCHLD;
Toy_Type *const_SIGUSR1;
Toy_Type *const_SIGUSR2;
Toy_Type *const_attrap;
Toy_Type *const_atscriptid;
Toy_Type *const_CWD;
Toy_Type *const_ENV;
Toy_Type *const_up;
Toy_Type *const_VERSION;
Toy_Type *const_nocase;
Toy_Type *const_all;
Toy_Type *const_string;
Toy_Type *const_delegate;
Toy_Type *const_arg;
Toy_Type *const_int1;
Toy_Type *const_new;
Toy_Type *const_unknown;

volatile int sig_flag;

static Toy_Type* parse_env(char* buff, char sep);


Toy_Interp* new_interp(char* name, int stack_size) {
    Toy_Interp *interp;

    const_Symbol = new_symbol("Symbol");
    const_Block = new_symbol("Block");
    const_List = new_symbol("List");
    const_Real = new_symbol("Real");
    const_Integer = new_symbol("Integer");
    const_String = new_symbol("String");
    const_Exception = new_symbol("Exception");
    const_RQuote = new_symbol("RQuote");
    const_Object = new_symbol("Object");
    const_Nil = new_nil();
    const_T = new_symbol("t");
    const_Question = new_symbol("?");
    const_Holder = new_symbol("@holder");
    const_Init = new_symbol("init");
    const_Hash = new_symbol("Hash");
    const_File = new_symbol("File");
    const_Args = new_symbol("args:");
    const_Array = new_symbol("Array");
    const_Get = new_symbol("get");
    const_atout = new_symbol("@out");
    const_atin = new_symbol("@in");
    const_stdin = new_symbol("stdin");
    const_stdout = new_symbol("stdout");
    const_stderr = new_symbol("stderr");
    const_filec = new_symbol("file:");
    const_gets = new_symbol("gets");
    const_puts = new_symbol("puts");
    const_atname = new_symbol("@name");
    const_atstacktrace = new_symbol("@stack-trace");
    const_SIGHUP = new_symbol("SIGHUP");
    const_SIGINT = new_symbol("SIGINT");
    const_SIGQUIT = new_symbol("SIGQUIT");
    const_SIGPIPE = new_symbol("SIGPIPE");
    const_SIGALRM = new_symbol("SIGALRM");
    const_SIGTERM = new_symbol("SIGTERM");
    const_SIGURG = new_symbol("SIGURG");
    const_SIGCHLD = new_symbol("SIGCHLD");
    const_SIGUSR1 = new_symbol("SIGUSR1");
    const_SIGUSR2 = new_symbol("SIGUSR2");
    const_attrap = new_symbol("@trap");
    const_atscriptid = new_symbol("@script-id");
    const_CWD = new_symbol("CWD");
    const_ENV = new_symbol("ENV");
    const_up = new_symbol("up:");
    const_VERSION = new_symbol("VERSION");
    const_nocase = new_symbol("nocase:");
    const_all = new_symbol("all:");
    const_string = new_symbol("string");
    const_delegate = new_symbol("delegate:");
    const_arg = new_symbol("arg:");
    const_int1 = new_integer(1);
    const_new = new_symbol("new");
    const_unknown = new_symbol("unknown");

    interp = GC_MALLOC(sizeof(Toy_Interp));
    if (NULL == interp) return NULL;

    interp->obj_stack = (Toy_Obj_Env**)(GC_MALLOC(sizeof(Toy_Obj_Env*) * stack_size));
    if (NULL == interp->obj_stack) return NULL;

    interp->func_stack = (Toy_Func_Env**)(GC_MALLOC(sizeof(Toy_Func_Env*) * stack_size));
    if (NULL == interp->func_stack) return NULL;

    interp->funcs = new_hash();
    if (NULL == interp->funcs) return NULL;

    interp->classes = new_hash();
    if (NULL == interp->classes) return NULL;

    interp->globals = new_hash();
    if (NULL == interp->globals) return NULL;

    interp->scripts = new_hash();
    if (NULL == interp->scripts) return NULL;

    interp->name = name;
    interp->stack_size = stack_size;
    interp->cur_obj_stack = -1;
    interp->cur_func_stack = -1;
    interp->line = 0;
    interp->statement = new_statement(new_list(new_symbol("(toysh)")), 0);
    interp->object_name = new_symbol("Toplevel");
    interp->func_name = new_symbol("Main");
    interp->script_id = 0;
    hash_set_t(interp->scripts, const_atscriptid, new_integer(interp->script_id));

    sig_flag = 0;

    return interp;
}

Toy_Interp* interp_setup(Toy_Interp* interp, int argc, char **argv, char **envp) {
    Toy_Type *delegate;
    Hash *gdict;
    int i;
    Toy_Type *argl, *l;
    Bulk *b;
    Toy_Type *any, *result;
    Toy_Type *env, *envl, *envv;
    Toy_Type *setupl;

    delegate = new_list(const_Object);

    toy_push_new_obj_env(interp, "Toplevel", new_object("Main", new_hash(), delegate));
    toy_push_func_env(interp, new_hash(), NULL);

    /* make world */

    toy_add_class(interp, "Functions", interp->funcs, delegate);
    // toy_add_class(interp, "Symbol", NULL, delegate);
    toy_add_class(interp, "Integer", NULL, delegate);
    toy_add_class(interp, "Real", NULL, delegate);
    toy_add_class(interp, "String", NULL, delegate);
    toy_add_class(interp, "List", NULL, delegate);
    toy_add_class(interp, "Block", NULL, delegate);
    // toy_add_class(interp, "Exception", NULL, delegate);
    toy_add_class(interp, "RQuote", NULL, delegate);
    toy_add_class(interp, "Object", NULL, NULL);
    // toy_add_class(interp, "Func", NULL, delegate);
    toy_add_class(interp, "Hash", NULL, delegate);
    toy_add_class(interp, "File", NULL, delegate);
    toy_add_class(interp, "Array", NULL, delegate);

    toy_add_commands(interp);
    toy_add_methods(interp);

    /* set VERSION */
    hash_set_t(interp->globals, const_VERSION, new_string_str(VERSION));

    /* set ARGV */
    l = argl = new_list(NULL);
    for (i=0; i<argc; i++) {
	l = list_append(l, new_string_str(argv[i]));
    }
    gdict = interp->globals;
    hash_set(gdict, "ARGV", argl);
    hash_set(gdict, "LIB_PATH", new_list(new_string_str(LIB_PATH)));

    /* set ENV */
    l = envl = new_list(new_symbol("new"));
    l = list_append(l, new_symbol("Hash"));
    env = toy_call(interp, envl);
    if (GET_TAG(env) == OBJECT) {
	for (; *envp; envp++) {
	    envv = parse_env(*envp, '=');
	    l = envl = new_list(env);
	    l = list_append(l, new_symbol("set"));
	    l = list_append(l, list_get_item(envv));
	    l = list_append(l, list_get_item(list_next(envv)));
	    toy_call(interp, envl);
	}
    }
    hash_set_t(interp->globals, const_ENV, env);

    cmd_pwd(interp, new_list(NULL), new_hash(), 0);

    /* load initial setup file "setup.toy" */
    setupl = new_list(new_symbol("load"));
    list_append(setupl, new_string_str(SETUP_FILE));
    any = toy_call(interp, setupl);
    if (GET_TAG(any) == EXCEPTION) {
	    fprintf(stderr, "Exception occurd in setup script file at load '%s', %s.\n",
		    SETUP_FILE, to_string(any));
    }

    return interp;
}

int toy_add_class(Toy_Interp* interp, char* name, Hash* slot, Toy_Type *delegate) {
    Toy_Type *obj;
    Hash *h;

    if (slot == NULL) {
	if (NULL == (h = new_hash())) return 0;
    } else {
	h = slot;
    }

    obj = new_object(name, h, delegate);
    if (NULL == hash_set(interp->classes, name, obj)) return 0;

    return 1;
}

static Toy_Type* parse_argspec(char* argspec);

void
toy_add_func(Toy_Interp* interp,
	     char* name,
	     Toy_Type* native(Toy_Interp*, Toy_Type*, Hash*, int),
	     char* argspec) {

    Toy_Type *argspec_list;
    Toy_Type *o;

    argspec_list = parse_argspec(argspec);
    o = new_native(native, argspec_list);

    hash_set(interp->funcs, name, o);
}

void
toy_add_method(Toy_Interp* interp,
	       char* class,
	       char* method,
	       Toy_Type* native(Toy_Interp*, Toy_Type*, Hash*, int),
	       char* argspec) {

    Toy_Type *argspec_list;
    Toy_Type *o;
    Toy_Type *h;

    h = hash_get(interp->classes, class);
    if (NULL == h) {
	assert(0);
    }

    argspec_list = parse_argspec(argspec);
    o = new_native(native, argspec_list);

    hash_set(h->u.hash, method, o);
}

static Toy_Type*
parse_argspec(char* argspec) {
    Toy_Type *l;
    Cell *c;
    char *p;

    if (argspec == NULL) return NULL;

    p = argspec;
    c = new_cell("");
    l = new_list(NULL);

    if (strlen(argspec) == 0) {
	list_append(l, const_Nil);
	return l;
    }

    while (*p) {
	if (*p == ',') {
	    list_append(l, new_symbol(cell_get_addr(c)));
	    c = new_cell("");
	} else {
	    cell_add_char(c, *p);
	}
	p++;
    }
    list_append(l, new_symbol(cell_get_addr(c)));

    return l;
}

static Toy_Type*
parse_env(char* buff, char sep) {
    Toy_Type *l;
    Cell *c;
    char *p;

    if (buff == NULL) return NULL;

    p = buff;
    c = new_cell("");
    l = new_list(NULL);

    if (strlen(buff) == 0) {
	list_append(l, const_Nil);
	return l;
    }

    while (*p) {
	if (*p == sep) {
	    list_append(l, new_string_str(cell_get_addr(c)));
	    c = new_cell("");
	    p++;
	    cell_add_str(c, p);
	    break;
	} else {
	    cell_add_char(c, *p);
	}
	p++;
    }
    list_append(l, new_string_str(cell_get_addr(c)));

    return l;
}

Toy_Obj_Env*
toy_new_obj_env(Toy_Interp *interp, Toy_Type *cur_object, Toy_Type *self) {
    Toy_Obj_Env *env;

    env = GC_MALLOC(sizeof(Toy_Obj_Env));
    if (NULL == env) return NULL;

    env->cur_object = cur_object;
    env->cur_object_slot = cur_object->u.object.slots;
    env->self = self;

    return env;
}

int
toy_push_new_obj_env(Toy_Interp* interp, char* class, Toy_Type* self) {
    Toy_Obj_Env *env;
    Toy_Type *o; 

    if ((interp->cur_obj_stack+1) >= interp->stack_size) return 0;

    o = new_object("Object", new_hash(), new_list(const_Object));
    if (NULL == (env = toy_new_obj_env(interp, o, o))) {
	return 0;
    }

    interp->cur_obj_stack++;
    interp->obj_stack[interp->cur_obj_stack] = env;

    hash_set_t(interp->obj_stack[interp->cur_obj_stack]->cur_object_slot,
	       const_atname, new_symbol(class));

    return 1;
}

int
toy_push_obj_env(Toy_Interp *interp, Toy_Obj_Env *env) {
    if ((interp->cur_obj_stack+1) >= interp->stack_size) return 0;

    interp->cur_obj_stack++;
    interp->obj_stack[interp->cur_obj_stack] = env;

    return 1;
}

Toy_Obj_Env*
toy_pop_obj_env(Toy_Interp* interp) {
    Toy_Obj_Env *env;

    env = interp->obj_stack[interp->cur_obj_stack];
    interp->obj_stack[interp->cur_obj_stack] = NULL;
    interp->cur_obj_stack--;

    return env;
}

int
toy_push_func_env(Toy_Interp* interp, Hash *localvar, Hash *closurevar) {
    Toy_Func_Env *env;

    if ((interp->cur_func_stack+1) >= interp->stack_size) return 0;

    env = GC_MALLOC(sizeof(Toy_Func_Env));
    if (NULL == env) return 0;

    env->localvar = localvar;
    if (localvar == closurevar) {
	env->closurevar = NULL;
    } else {
	env->closurevar = closurevar;
    }
    env->statement = interp->statement;
    env->line = interp->line;
    env->script_id = interp->script_id;
    env->object_name = interp->object_name;
    env->func_name = interp->func_name;

    interp->cur_func_stack++;
    interp->func_stack[interp->cur_func_stack] = env;

    return 1;
}

Toy_Func_Env*
toy_pop_func_env(Toy_Interp* interp) {
    Toy_Func_Env *env;

    env = interp->func_stack[interp->cur_func_stack];
    interp->func_stack[interp->cur_func_stack] = NULL;
    interp->cur_func_stack--;

    return env;
}

Toy_Env*
new_closure_env(Toy_Interp* interp) {
    Toy_Env *env;

    if (NULL == (env = GC_MALLOC(sizeof(Toy_Env)))) return NULL;

    env->localvar = interp->func_stack[interp->cur_func_stack]->localvar;
    env->object_env = interp->obj_stack[interp->cur_obj_stack];

    return env;
}


static char *get_script_path(Toy_Interp* interp, int id);

char*
get_stack_trace(Toy_Interp *interp, Cell *stack) {
    char buff[256];
    int i;

    snprintf(buff, 256, "%s:%d: %s in %s::%s\n",
	     get_script_path(interp, interp->func_stack[interp->cur_func_stack]->script_id),
	     interp->line,
	     to_print(interp->statement),
	     (interp->func_stack[interp->cur_func_stack]->object_name ?
	      		to_string(interp->func_stack[interp->cur_func_stack]->object_name) : "self"),
	     (interp->func_stack[interp->cur_func_stack]->func_name ?
	      		to_string(interp->func_stack[interp->cur_func_stack]->func_name) : "???"));
    buff[255] = 0;
    cell_add_str(stack, buff);

    for (i=interp->cur_func_stack; i>0; i--) {
	snprintf(buff, 256, "%s:%d: %s in %s::%s\n",
		 get_script_path(interp, interp->func_stack[i-1]->script_id),
		 interp->func_stack[i]->line,
		 to_print(interp->func_stack[i]->statement),
		 (interp->func_stack[i-1]->object_name ?
		  		to_string(interp->func_stack[i-1]->object_name) : "self"),
		 (interp->func_stack[i-1]->func_name ?
		  		to_string(interp->func_stack[i-1]->func_name) : "???"));
	buff[255] = 0;
	cell_add_str(stack, buff);
    }

    return cell_get_addr(stack);
}

static char*
get_script_path(Toy_Interp* interp, int script_id) {
    Hash *h;
    Toy_Type *cont;
    Toy_Script *s;

    if (0 == script_id) return "(no_source)";

    h = interp->scripts;
    cont = hash_get(h, to_string(new_integer(script_id)));
    if (NULL == cont) return "???";

    s = (Toy_Script*)cont->u.container;
    return cell_get_addr(s->path->u.string);
}
